from casadi import *

# Task 3.1
R = DMatrix(3,3)
R[0,0] = sqrt(2)
R[2,2] = sqrt(2*100.)
print 'R = ', R

# Task 3.2
x = SX.sym('x',3)
g = x[2] + (1-x[0])**2 - x[1]
G = SXFunction([x],[g])
G.init()
x_test = [2.,3.,4.]
[g_test] = G([x_test])
print 'g(x_test) = ', g_test

# Task 3.3
A = jacobian(g,x)
Jg = SXFunction([x], [A])
Jg.init()
x_test = [2.,3.,4.]
[Jg_test] = Jg([x_test])
print 'Jg(x_test) = ', Jg_test

# Task 3.4
#   minimize(dx)     1/2  dx' I dx
#   subject to
#                    -g(xk) <= Jg dx <= -g(xk)

# Task 3.5

# Allocate a QP solver
H = mul(R.T, R)
qp = qpStruct(h=H.sparsity(), a=A.sparsity())
solver = QpSolver("nlp.ipopt",qp)   # Allocate solver (IPOPT)
#solver = QpSolver("qpoases",qp)    # Allocate solver (qpOASES)
solver.init()                       # Initialize the solver

# SQP iterations
xk = DMatrix([2.5, 3.0, 0.75])
xk_all = [xk]

# Primal and dual infeasibility
inf_pr = []
inf_du = []

for k in range(5):
    # Obtain QP
    [gk] = G([xk])
    [Ak] = Jg([xk])

    # Solver the QP
    solver.setInput(H,"h")
    solver.setInput(mul(R.T,mul(R,xk)),"g")
    solver.setInput(Ak,"a")
    solver.setInput(-gk,"lba")
    solver.setInput(-gk,"uba")
    solver.evaluate()
    dxk = solver.getOutput("x")
    lam_gk = solver.getOutput("lam_a")

    # Check convergence
    inf_pr_k = gk
    inf_du_k = mul(R, xk) + mul(Ak.T, lam_gk)
    inf_pr.append(norm_2(inf_pr_k))
    inf_du.append(norm_2(inf_du_k))

    # Take the step
    xk = xk + dxk
    xk_all.append(xk)

for (k,xk) in enumerate(xk_all):
    print 'Iteration %3s: %20s' % (k, xk)

# Show convergence
x0_all = [xk[0] for xk in xk_all]
x1_all = [xk[1] for xk in xk_all]
x2_all = [xk[2] for xk in xk_all]

from matplotlib import pylab as plt
plt.clf()
plt.subplot(2,1,1)
plt.hold(True)
plt.grid(True)
plt.plot(x0_all)
plt.plot(x1_all)
plt.plot(x2_all)
plt.legend(('x0', 'x1', 'x2'), loc='lower right')

plt.subplot(2,1,2)
plt.hold(True)
plt.grid(True)
plt.semilogy(inf_pr)
plt.semilogy(inf_du)
plt.legend(('Primal infeasibility', 'Dual infeasibility'), loc='upper right')

plt.show()

